// Interface zwischen bis zu acht MERG Handsets MK2 und DCCpp_Basestation
// Author: Harald Sattler
// MK2 sammelt mehrere Tasten und Schalter sowie die Handset-Adresse (0..99) in ein serielles Telegramm.
// Das Interface taktet diese Informationen aus dem MK2 heraus und leitet die gewonnene Information 
// an die DCC++ Basestation als ASCII-Befehl weiter. Notation siehe Beschreibung DCC-Basestation.
// Die Handsetadresse wird als DCC LokAdresse verwendet, die Loks auf dem Layout muessen also Adressen 
// im Bereich 0..99 erhalten.
// Die Taste "Consist set" am Handset schickt ein NotHalt ("<t1 3 -1 1>") an die zugehrige Lok (hier Handset 1, Adr. 3)
//
// Informationen zum Handset MK2 von MERG
// Aufbau der seriellen Information vom MK2                                 __Handset__
// Bit  Funktion    | Kommentar                                            |           |
//  0   Consist set | serial IN                          On/Off  ->        #           #        <- Fwd/Rev
//  1   Adresse 1er | HC165 - 1                                            |        O  |        <- Poti
//  2   . BCD                                            F0      ->        |  O        |
//  3   .  -                                             F1      ->        |  O    * * |        <- Adress-Schalter
//  4   .   kodiert                                      F2      ->        |  O        |
//  5   Adresse 10er                                     F3      ->        |  O      O |        <- Nothalt
//  6   . BCD                                                              |____###____|
//  7   .  -                                                                 Anschluss		
//  8   .   kodiert
//  9   F0           | HC165 - 2
//  10  F1
//  11  F2
//  12  F3
//  13  FWD / REV
//  14  SpeedSteps
//  15  SpeedSteps
//  16  Off / On
//
// Anmerkungen zu den Analogeingaengen:
// Die Verwendung der AnalogScanner Bibliothek verhindert den Betrieb der SoftwareSerial Bibliothek.
// Letztere wird benoetigt, um die DCC++ Befehle an den Arduino Mega2560 (DCC++ BaseStation) zu senden.
// Es existiert eine weitere SoftwareSerial Implementierung die _angeblich_ mit anderen Interrrupt-gesteuerten 
// Libraries kooperiert: AltSoftSerial von Paul Stoffregen
// Allerdings behauptet das nur die Kurzbeschreibung im Arduino LibManager, die Homepage sagt etwas ganz anderes:
// Nur maximal 31250 Baud @16 MHz, Interferenzen mit anderen Int-Prozessen, Timer2 wird exklusiv bentigt, ...
// Somit eher ungeeignet :(
//
// Damit die BaseStation die Befehle der MK2 Handsets vom Interface empfangen kann, muss in der Datei "SerialCommand.cpp" 
// in der Routine "void SerialCommand::process()" die Abfrage der 2. Seriellen SS "Serial1" des Arduino Mega2560 implementiert werden.
// In der Datei "DCCpp_Uno.ino" muss die 2. serielle SS aktiviert werden: In setup() "Serial1.begin(115200);".
//
// 15.04.2019
// - Umstellung der Array-Zugriffe auf sprechende Makros zur besseren Lesbarkeit des Code
// 16.04.2019 - V2
// - Umstellung des Handset-Scan auf Interrupt um unterschiedlich lange Tastendruecke erkennbar zu machen.
// - Hardware-Serial fuer die Kommunikation zur DCC++ BaseStation verwenden, da die SoftSerial nicht zuverlaessig laeuft.
//	 Aktivierung der Umschaltung ueber #define HWSerialDCC.
//	 Ausgaben, die ohne die Umschaltung ueber die HW-Serielle und USB an den PC gehen, werden unterdrueckt.
//	 Die Ausgaben an die BaseStation werden parallel ueber USB auch an den Rechner gesendet.
// 17.04.2019
// - F0 an einem der Handsets lange druecken schaltet die DCC Spannung ein/aus
// - NH am Handset druecken loest einen Nothalt der betreffenden Lok aus (ohne Bremsverzoegerung, direkt Stop)
// - NH an einem Handset laenger als 0,5s gedrueckt halten loest einen Nothalt ALLER Loks aus
//   Nach einem Nothalt muessen die Potis der Handsets zuerst wieder auf Null gestellt werden, bevor eine Lok wieder fahren darf
// - Die Tasten NH und F0 werden auch ausgewertet (und an die BaseStation gesendet) wenn der Handset-Schalter auf OFF steht. 
//   Somit kann an jedem Handset der Nothalt ausgeloest oder die DCC Spannung ein/aus geschaltet werden, auch wenn dieses Handset gerade nicht benutzt wird.
// - Nach PowerOn stehen alle Handsets auf Nothalt -> Poti muss erst auf Null gestellt werden, bevor die Lok fahren darf
// - Wenn ein Handset abgeschaltet wird, wird Status "Nothalt" fuer diese Lok gesetzt -> Poti muss anschliessend Null sein, bevor die Lok wieder fahren darf
// 18.04.2019 - V2.0
// - Umschalten der Fahrtrichtung muss einen Fahrbefehl ausloesen, sonst kommt der Richtungswechsel nicht bei der Lok an
// - SoftwareSerial ausgebaut, kann sowieso nicht verwendet werden (ungeklaerte Ursache - bei bestimmten Software-Konstellationen
//   oder wenn der Code zu umfangreich wird, funktioniert SoftSerial nicht mehr)
// 21.04.2019 - V2.1
// - Taster F0 umbenannt in NH (Nothalt), Taster F1 wird F0, F2 wird F1, F3 wird F2, F4 wird F3
//   Taster F0 sendet 144 (statt 129) -> Licht am LokDecoder
// 29.04.2019 - V2.2
// - Die Tasten F0..F3 wirken nur momentan, z.B. das Licht ist nur solange an, wie die Taste F0 gedrueckt ist -> als Toggle-Schalter implementiert
//   Die Funktion der Tasten F0..F3 als Toggle oder Momentan wird per serielle SS je Taste festgelegt und im EEPROM hinterlegt
//   Mit Eingabe von "$$" ueber die serielle Schnittstelle wird ausgegeben, wie die taster momentan konfiguriert sind (Taster oder Schalter)
//   Durch Eingabe von "Fx=T" bzw. "Fx=S" wird die Taste Fx als Taste bzw. als Schalter definiert, x=0..3 
//   "Toggle" bedeutet, beim Druck auf die Taste wird deren Code einmalig an die BaseStation gesendet, beim Loslassen wird NICHT wieder der 
//	 Code fuer "keine Taste gedrueckt" gesendet, das passiert erst wieder beim erneuten Druck auf dieselbe Taste.
//   Genaugenommen wird der jeweils letzte Tastenzustand gespeichert und bei Toggle-Tasten (Schaltern) beim zweiten Druck dieser Zustand wiederhergestellt.

/*
This code is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this code; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110 - 1301  USA
*/

// nachfolgende Zeile einkommentieren um die loop-Zeit mit dem Oszi an LED13 zu messen
//#define TIMINGCHECK

#include <MsTimer2.h>
#include <eeprom.h>

// Version und Datum
#define MajorVersion 2
#define MinorVersion 2
#define BuildDate "29.04.2019"


// PIN Definitionen
#define MK2_Poti1 A0			// Analogeingang fuer Speed-Poti
#define MK2_Poti2 A1			// Analogeingang fuer Speed-Poti
#define MK2_Poti3 A2			// Analogeingang fuer Speed-Poti
#define MK2_Poti4 A3			// Analogeingang fuer Speed-Poti
#define MK2_Poti5 A4			// Analogeingang fuer Speed-Poti
#define MK2_Poti6 A5			// Analogeingang fuer Speed-Poti
#define MK2_Poti7 A6			// Analogeingang fuer Speed-Poti
#define MK2_Poti8 A7			// Analogeingang fuer Speed-Poti
#define MK2_CLKout 12			// Clockimpulse zum Einlesen der seriellen Daten vom MK2
#define MK2_DATAin1 9			// serielle Daten vom MK2 - Lok 1
#define MK2_DATAin2 8			// serielle Daten vom MK2 - Lok 2
#define MK2_DATAin3 7			// serielle Daten vom MK2 - Lok 3
#define MK2_DATAin4 6			// serielle Daten vom MK2 - Lok 4
#define MK2_DATAin5 5			// serielle Daten vom MK2 - Lok 5
#define MK2_DATAin6 4			// serielle Daten vom MK2 - Lok 6
#define MK2_DATAin7 3			// serielle Daten vom MK2 - Lok 7
#define MK2_DATAin8 2			// serielle Daten vom MK2 - Lok 8
#define SignalLED 13			// Anzeige von Kommunikation

#define sampleInterval 10		// alle "sampleInterval" ms die Schalter im MK2 abfragen
								// nicht zu oft, weil der ParallelLoad fuer die 74HC165 mit einem 74HC123 erzeugt wird, dessen Zeitkonstante ~670 micros ist

// globale Variablen - allgemein
unsigned long loopTiming;		// Timer fuer Auslesen der Schalter der Handsets
unsigned long LEDmillis;		// Timer fuer die LED, Anzeige von Kommunikation (aenderung von Geschwindigkeit oder Schalterstellung einer der Loks)
volatile unsigned char LokNummer=0;		// welche Lok wird gerade in der Loop bearbeitet
bool DCCSpannungEin = false;	// DCC Anlagenspannung ein/aus ueber "Taste F1 lange druecken" an jedem Handset - Default: AUS
bool NotHaltALLE = false;		// F0 lange gedrueckt, ALLE Loks anhalten - Status merken und Befehl mehrfach hintereinander abgeben um Kontaktprobleme auszutricksen
unsigned long uNotAusZeit;		// "NotHalt fuer ALLE" muss mehrfach mit zeitlichem Abstand ausgegeben werden
byte NotHaltALLECounter = 0;	// "NotHalt fuer ALLE" wird 10 mal ausgegeben
String inputString = "";        // Speicher fr Eingabestrings auf der seriellen SS
boolean stringComplete = false; // CR/LF empfangen
unsigned int functionCode;		// F-Tasten Code fuer Uebertragung an DCC++ BaseStation

// globale Variablen - je Lok	(wohl am besten jeweils in einer struct ablegen zum Interieren)
volatile unsigned long SwVar1, SwVar2, SwVar3, SwVar4, SwVar5, SwVar6, SwVar7, SwVar8;		// die einzelnen Schalter-Bits in einem Long aufgesammelt
unsigned int  SwVar1Int, SwVar2Int, SwVar3Int, SwVar4Int, SwVar5Int, SwVar6Int, SwVar7Int, SwVar8Int;	// Schalter-Bits auf Integer abgebildet
unsigned long SwVar1Old=0, SwVar2Old = 0, SwVar3Old = 0, SwVar4Old = 0, SwVar5Old = 0, SwVar6Old = 0, SwVar7Old = 0, SwVar8Old = 0;	// letzter Zustand der Schalter
unsigned int Adresse1, Adresse2, Adresse3, Adresse4, Adresse5, Adresse6, Adresse7, Adresse8;	// Lok Adresse
unsigned long PotiLok1,    PotiLok2,    PotiLok3,    PotiLok4,    PotiLok5,    PotiLok6,    PotiLok7,    PotiLok8;		// Potivalue Handset (0..1023)
unsigned int SpeedLok1Old, SpeedLok2Old, SpeedLok3Old, SpeedLok4Old, SpeedLok5Old, SpeedLok6Old, SpeedLok7Old, SpeedLok8Old;	// vorheriger Speedvalue Lok (0..14, 0..28, 0..126, je nach ModusBits)
unsigned int Speed1, Speed2, Speed3, Speed4, Speed5, Speed6, Speed7, Speed8;			// auf SpeedSteps (14, 28 oder 128) gemappte Lokgeschwindigkeit
byte Lok1Register=1, Lok2Register=2, Lok3Register = 3, Lok4Register = 4, Lok5Register = 5, Lok6Register = 6, Lok7Register = 7, Lok8Register = 8; // DCC++ intern verwendeter Register Wert je Lok
byte ModusBits1=3, ModusBits2 = 3, ModusBits3 = 3, ModusBits4 = 3, ModusBits5 = 3, ModusBits6 = 3, ModusBits7 = 3, ModusBits8 = 3;				// Fahrstufen 14, 18 oder 128, kodiert

//--->>> hier Var auf char umstellen und Flag fr triggered plus die per F-Taste gesetzten Bits speichern, im Code dann ausmaskieren fr speichern und auswerten
byte iFxTrig1, iFxTrig2, iFxTrig3, iFxTrig4, iFxTrig5, iFxTrig6, iFxTrig7, iFxTrig8;	// Merker fuer Tastendruck F0..F3
bool bF0Long1, bF0Long2, bF0Long3, bF0Long4, bF0Long5, bF0Long6, bF0Long7, bF0Long8;	// Merker fuer langen Tastendruck F0
unsigned long lF0time1, lF0time2, lF0time3, lF0time4, lF0time5, lF0time6, lF0time7, lF0time8;	// Zeitpunkt Tastendruck F0
bool bNHTrig1, bNHTrig2, bNHTrig3, bNHTrig4, bNHTrig5, bNHTrig6, bNHTrig7, bNHTrig8;	// Merker fuer Tastendruck NH
bool bNHLong1, bNHLong2, bNHLong3, bNHLong4, bNHLong5, bNHLong6, bNHLong7, bNHLong8;	// Merker fuer langen Tastendruck NH
unsigned long lNHtime1, lNHtime2, lNHtime3, lNHtime4, lNHtime5, lNHtime6, lNHtime7, lNHtime8;	// Zeitpunkt Tastendruck NH
bool bNothalt1, bNothalt2, bNothalt3, bNothalt4, bNothalt5, bNothalt6, bNothalt7, bNothalt8;	// Merker, dass NotHalt fuer die Lok ausgeloest wurde (dann muss Speed auf Null gestellt werden, bevor Speed wieder gesetzt werden kann)
bool bLokDir1Old, bLokDir2Old, bLokDir3Old, bLokDir4Old, bLokDir5Old, bLokDir6Old, bLokDir7Old, bLokDir8Old;	// Merker fuer aktuelle Fahrtrichtung der Lok
byte bFxOld1, bFxOld2, bFxOld3, bFxOld4, bFxOld5, bFxOld6, bFxOld7, bFxOld8;			// Merker, ob Schalter einer F-Taste On oder Off ist

volatile unsigned long LokArray[8][20] = {									// 8 Loks mit je 20 Variablen
	//     0      ,   1   ,     2    ,     3    ,     4   ,     5   ,       6     ,    7  ,       8    ,     9    ,     10    ,    11   ,    12   ,    13   ,    14   ,    15   ,    16   ,    17    ,     18     ,    19
	{ Lok1Register, SwVar1, SwVar1Int, SwVar1Old, Adresse1, PotiLok1, SpeedLok1Old, Speed1, MK2_DATAin1, MK2_Poti1, ModusBits1, iFxTrig1, bF0Long1, bNHTrig1, bNHLong1, lF0time1, lNHtime1, bNothalt1, bLokDir1Old, bFxOld1 },	// Lok 1
	{ Lok2Register, SwVar2, SwVar2Int, SwVar2Old, Adresse2, PotiLok2, SpeedLok2Old, Speed2, MK2_DATAin2, MK2_Poti2, ModusBits2, iFxTrig2, bF0Long2, bNHTrig2, bNHLong2, lF0time2, lNHtime2, bNothalt2, bLokDir2Old, bFxOld2 },	// Lok 2
	{ Lok3Register, SwVar3, SwVar3Int, SwVar3Old, Adresse3, PotiLok3, SpeedLok3Old, Speed3, MK2_DATAin3, MK2_Poti3, ModusBits3, iFxTrig3, bF0Long3, bNHTrig3, bNHLong3, lF0time3, lNHtime3, bNothalt3, bLokDir3Old, bFxOld3 },	// Lok 3
	{ Lok4Register, SwVar4, SwVar4Int, SwVar4Old, Adresse4, PotiLok4, SpeedLok4Old, Speed4, MK2_DATAin4, MK2_Poti4, ModusBits4, iFxTrig4, bF0Long4, bNHTrig4, bNHLong4, lF0time4, lNHtime4, bNothalt4, bLokDir4Old, bFxOld4 },	// Lok 4
	{ Lok5Register, SwVar5, SwVar5Int, SwVar5Old, Adresse5, PotiLok5, SpeedLok5Old, Speed5, MK2_DATAin5, MK2_Poti5, ModusBits5, iFxTrig5, bF0Long5, bNHTrig5, bNHLong5, lF0time5, lNHtime5, bNothalt5, bLokDir5Old, bFxOld5 },	// Lok 5
	{ Lok6Register, SwVar6, SwVar6Int, SwVar6Old, Adresse6, PotiLok6, SpeedLok6Old, Speed6, MK2_DATAin6, MK2_Poti6, ModusBits6, iFxTrig6, bF0Long6, bNHTrig6, bNHLong6, lF0time6, lNHtime6, bNothalt6, bLokDir6Old, bFxOld6 },	// Lok 6
	{ Lok7Register, SwVar7, SwVar7Int, SwVar7Old, Adresse7, PotiLok7, SpeedLok7Old, Speed7, MK2_DATAin7, MK2_Poti7, ModusBits7, iFxTrig7, bF0Long7, bNHTrig7, bNHLong7, lF0time7, lNHtime7, bNothalt7, bLokDir7Old, bFxOld7 },	// Lok 7
	{ Lok8Register, SwVar8, SwVar8Int, SwVar8Old, Adresse8, PotiLok8, SpeedLok8Old, Speed8, MK2_DATAin8, MK2_Poti8, ModusBits8, iFxTrig8, bF0Long8, bNHTrig8, bNHLong8, lF0time8, lNHtime8, bNothalt8, bLokDir8Old, bFxOld8 }	// Lok 8
};

// defines fuer sprechenden Zugriff auf das LokArray
#define LOKREG LokArray[LokNummer][0]
#define SWVAR LokArray[LokNummer][1]
#define SWVARINT LokArray[LokNummer][2]
#define SWVAROLD LokArray[LokNummer][3]
#define LOKADR LokArray[LokNummer][4]
#define LOKPOTI LokArray[LokNummer][5]
#define LOKSPEEDOLD LokArray[LokNummer][6]
#define LOKSPEED LokArray[LokNummer][7]
#define MK2DATAIN LokArray[LokNummer][8]
#define MK2POTIIN LokArray[LokNummer][9]
#define MODUSBITS LokArray[LokNummer][10]

#define FxTRIG LokArray[LokNummer][11]
#define F0LONG LokArray[LokNummer][12]
#define NHTRIG LokArray[LokNummer][13]
#define NHLONG LokArray[LokNummer][14]
#define F0TIME LokArray[LokNummer][15]
#define NHTIME LokArray[LokNummer][16]
#define NOTHALT LokArray[LokNummer][17]
#define LOKDIROLD LokArray[LokNummer][18]
#define FxOLD LokArray[LokNummer][19]

// defines fr Parameter in/aus EEPROM - EEPROM Parameter gelten fuer ALLE Loks
#define EEF0 100															// EEPROM Adresse fuer Parameter "Taste F0 ist Taster/Schalter"
#define EEF1 102															// EEPROM Adresse fuer Parameter "Taste F1 ist Taster/Schalter"
#define EEF2 104															// EEPROM Adresse fuer Parameter "Taste F2 ist Taster/Schalter"
#define EEF3 106															// EEPROM Adresse fuer Parameter "Taste F3 ist Taster/Schalter"
#define TASTER 1															// Flag fuer Taster
#define SCHALTER 2															// Flag fuer Taster

// defines fuer Tasten innerhalb des Bitfelds FxTRIG (byte Feld)
#define GETFxF0 ((FxTRIG & 0x01) ? true : false)				//  == 0x01
#define GETFxF1 ((FxTRIG & 0x02) ? true : false)				//  == 0x02
#define GETFxF2 ((FxTRIG & 0x04) ? true : false)				//  == 0x04
#define GETFxF3 ((FxTRIG & 0x08) ? true : false)				//  == 0x08
#define GETFxNH ((FxTRIG & 0x10) ? true : false)				//  == 0x10

#define SETFxF0(x) if((x) == true) {(FxTRIG |= 0x01);} else {(FxTRIG &= ~(0x01));} 
#define SETFxF1(x) if((x) == true) {(FxTRIG |= 0x02);} else {(FxTRIG &= ~(0x02));}
#define SETFxF2(x) if((x) == true) {(FxTRIG |= 0x04);} else {(FxTRIG &= ~(0x04));}
#define SETFxF3(x) if((x) == true) {(FxTRIG |= 0x08);} else {(FxTRIG &= ~(0x08));}

// defines fuer Tasten innerhalb des Bitfelds FxOLD (byte Feld)
#define GETFxOldF0 ((FxOLD & 0x01) ? true : false)				//  == 0x01
#define GETFxOldF1 ((FxOLD & 0x02) ? true : false)				//  == 0x02
#define GETFxOldF2 ((FxOLD & 0x04) ? true : false)				//  == 0x04
#define GETFxOldF3 ((FxOLD & 0x08) ? true : false)				//  == 0x08

#define SETFxOldF0(x) ((x)) ? (FxOLD |= 0x01) : (FxOLD &= ~(0x01))
#define SETFxOldF1(x) ((x)) ? (FxOLD |= 0x02) : (FxOLD &= ~(0x02))
#define SETFxOldF2(x) ((x)) ? (FxOLD |= 0x04) : (FxOLD &= ~(0x04))
#define SETFxOldF3(x) ((x)) ? (FxOLD |= 0x08) : (FxOLD &= ~(0x08))
#define SETFxOldNH(x) ((x)) ? (FxOLD |= 0x10) : (FxOLD &= ~(0x10))

// defines fuer Merker Schalter/Taster im Bitfeld FxOLD (byte Feld) - Bit gesetzt -> Taster
#define GETTSF0 ((FxOLD & 0x10) ? TASTER : SCHALTER)				//  == 0x10
#define GETTSF1 ((FxOLD & 0x20) ? TASTER : SCHALTER)				//  == 0x20
#define GETTSF2 ((FxOLD & 0x40) ? TASTER : SCHALTER)				//  == 0x40
#define GETTSF3 ((FxOLD & 0x80) ? TASTER : SCHALTER)				//  == 0x80

#define SETTSF0(x) ((x == true)) ? (FxOLD |= 0x10) : (FxOLD &= ~(0x10))
#define SETTSF1(x) ((x == true)) ? (FxOLD |= 0x20) : (FxOLD &= ~(0x20))
#define SETTSF2(x) ((x == true)) ? (FxOLD |= 0x40) : (FxOLD &= ~(0x40))
#define SETTSF3(x) ((x == true)) ? (FxOLD |= 0x80) : (FxOLD &= ~(0x80))

// defines fuer Tastenabfrage (invertierte Logik, Tasten ziehen Signal auf LOW)	
#define GETKEYNH ((LokArray[LokNummer][1] & 0x00001 ) ? 0 : 1)			// == 0x00001		
#define GETKEYF0 ((LokArray[LokNummer][1] & 0x00200 ) ? 0 : 1)			// == 0x00200		
#define GETKEYF1 ((LokArray[LokNummer][1] & 0x00400 ) ? 0 : 1)			// == 0x00400		
#define GETKEYF2 ((LokArray[LokNummer][1] & 0x00800 ) ? 0 : 1)			// == 0x00800		
#define GETKEYF3 ((LokArray[LokNummer][1] & 0x01000 ) ? 0 : 1)			// == 0x01000		
#define GETMODUS ((LokArray[LokNummer][1] & 0x0C000 ) ? 0 : 1)			// == 0x0C000		
#define GETHSON	 ((LokArray[LokNummer][1] & 0x10000 ) ? 0 : 1)			// == 0x10000
#define LOKDIR	 ((LokArray[LokNummer][1] & 0x02000 ) ? 1 : 0)			// == 0x02000

// ISR fuer MsTimer2 ----------------------------------------------------------------------------------------------------------------
// Alle 1 ms wird ein Burst von 17 Impulsen von zusammen ~250 Microsekunden Laenge ausgeloest, der die Daten aus dem Handset in den Arduino taktet.
// In den Burst-Pausen wird vom 74HC123 der Parallel-Load Impuls erzeugt, der die parallelen Daten der Schalter in die Schieberegister 74HC168 einliest.
void getMK2Bits()
{
	static char lokalLokNummer;												// muss static deklariert sein, damit der Zaehler funktioniert
	lokalLokNummer = (lokalLokNummer + 1) % 8;								// Loknummer bei jedem Durchlauf inkrementieren -> LokNummer: 0..7

	unsigned long SchalterVariable = 0;										// definierter Aufsetzpunkt (alle Bits geloescht)

	for (int i = 0; i < 17; i++)											// MK2 hat 17 Bits fuer Schalter und Adressen
	{
		digitalWrite(MK2_CLKout, 1);										// CLK Impuls EIN (min 17..100 ns)
		digitalWrite(MK2_CLKout, 0);										// CLK Impuls wieder AUS

		SchalterVariable |= ((digitalRead(LokArray[lokalLokNummer][8]) & 0x01));			// binaer verODERt mit LSB
		(i < 16) ? SchalterVariable = SchalterVariable << 1 : 1 == 1;		// die letzte Schiebeoperation unterdruecken damit die 16 bit im Wort bleiben

	}
	LokArray[lokalLokNummer][1] = SchalterVariable;

#ifdef TIMINGCHECK
	//digitalWrite(SignalLED, !digitalRead(SignalLED));						// toggle LED fuer Laufzeitmessung
#endif
}

// setup() --------------------------------------------------------------------------------------------------------------------------
void setup() {
// set data rate for the Serial port
	Serial.begin(115200);
	inputString.reserve(20);												// 20 byte fuer inputString reservieren

// Parameter im EEPROM setzen, falls noch nicht erfolgt (Flag ist weder TASTER noch SCHALTER)
	if ((eepromReadInt(EEF0) != TASTER) && (eepromReadInt(EEF0) != SCHALTER))	eepromWriteInt(EEF0, TASTER);	// vorbelegen mit Taster
	if ((eepromReadInt(EEF1) != TASTER) && (eepromReadInt(EEF1) != SCHALTER))	eepromWriteInt(EEF1, TASTER);	// vorbelegen mit Taster
	if ((eepromReadInt(EEF2) != TASTER) && (eepromReadInt(EEF2) != SCHALTER))	eepromWriteInt(EEF2, TASTER);	// vorbelegen mit Taster
	if ((eepromReadInt(EEF3) != TASTER) && (eepromReadInt(EEF3) != SCHALTER))	eepromWriteInt(EEF3, TASTER);	// vorbelegen mit Taster

									

// Taster-Flags aus EEPROM ins RAM (LokArray[LokNummer][19]) laden
	if (eepromReadInt(EEF0) == TASTER) SETTSF0(true); else { SETTSF0(false); SETFxOldF0(false); }	// F0 ist als Taster parametriert? // bei Schalter: Toggle-Merker auf false, damit beim naechsten Druck auf Taster EIN-geschaltet wird
	if (eepromReadInt(EEF1) == TASTER) SETTSF1(true); else SETTSF1(false); 	// F1 ist als Taster parametriert?
	if (eepromReadInt(EEF2) == TASTER) SETTSF2(true); else SETTSF2(false); 	// F2 ist als Taster parametriert?
	if (eepromReadInt(EEF3) == TASTER) SETTSF3(true); else SETTSF3(false); 	// F3 ist als Taster parametriert?

	functionCode = 128;														// der Taster-Code ist immer mindestens 128, je nach aktiver F-Taste wird er erhoeht

// IOs initialisieren
	pinMode(MK2_CLKout, OUTPUT);											// Clock-Ausgang fuer die Handsets
	pinMode(SignalLED, OUTPUT);												// signalisiert fuer 100 ms, wenn Daten an die BaseStation uebertragen werden
	pinMode(MK2_DATAin1, INPUT_PULLUP);										// 8 digtale Eingaenge fuer die Schalter und Taster
	pinMode(MK2_DATAin2, INPUT_PULLUP);
	pinMode(MK2_DATAin3, INPUT_PULLUP);
	pinMode(MK2_DATAin4, INPUT_PULLUP);
	pinMode(MK2_DATAin5, INPUT_PULLUP);
	pinMode(MK2_DATAin6, INPUT_PULLUP);
	pinMode(MK2_DATAin7, INPUT_PULLUP);
	pinMode(MK2_DATAin8, INPUT_PULLUP);
	pinMode(MK2_Poti1, INPUT);												// 8 ananloge Eingaenge fuer die Potis
	pinMode(MK2_Poti2, INPUT);
	pinMode(MK2_Poti3, INPUT);
	pinMode(MK2_Poti4, INPUT);
	pinMode(MK2_Poti5, INPUT);
	pinMode(MK2_Poti6, INPUT);
	pinMode(MK2_Poti7, INPUT);
	pinMode(MK2_Poti8, INPUT);

	// damit wird sichergestellt, dass die LokPotis auf Null stehen, bevor nach dem Einschalten eine Geschwindigkeit an die zugehoerige Lok gesendet wird
	for (byte i = 0; i <= 7; i++)											// NotHalt Flag fuer alle Loks setzen
	{
		LokArray[i][17] = true;
	}

// Timer2 initialisieren und starten - in der Timer ISR werden alle Tasten und Schalter der Handsets ausgelesen und in das LokArray geschrieben
	MsTimer2::set(1, getMK2Bits);											// wird alle 1 ms aufgerufen
	MsTimer2::start();														// Timer2 starten

	loopTiming = millis();													// LoopTiming initialisieren
}

// loop() ---------------------------------------------------------------------------------------------------------------------------
void loop() {
	static byte n = 2;														// Anzahl virtueller Mittelwerte fuer Poti einlesen (Glaettung der Werte gegen Zittern)
																			// 2 arbeitet gut, 1 ist fuer 2 Handsets auch noch ok 
																			// ab 3 Handsets besser 2 - maximal jedoch 4, da sonst die Nachfuehrung der Werte zu lange dauert
#ifdef TIMINGCHECK
	//digitalWrite(SignalLED, !digitalRead(SignalLED));	// toggle LED fuer Laufzeitmessung
	digitalWrite(SignalLED, HIGH);						// LED ein fuer Laufzeitmessung
#endif
	//Serial.write(".");							// DEBUG

#ifndef TIMINGCHECK
	if (millis() > LEDmillis + 100) digitalWrite(SignalLED, 0);				// nach 100 ms die LED immer ausschalten
#endif

	if (millis() > loopTiming + sampleInterval)								// sampleInterval = 10 ms
	{
		loopTiming = millis();
		LokNummer = (LokNummer + 1) % 8;									// Loknummer bei jedem Durchlauf inkrementieren -> LokNummer: 0..7
		//Serial.print(LokNummer);	Serial.print(" ");							// Debug

		// hier die Schalter auswerten ------------------------------		// die Schalter werden per IRQ Routine eingelesen

		// Behandlung NOTHALT mit Taste "Consist" (== "NH")					// NH wird auch mit ausgeschaltetem Handset abgefragt und ausgewertet
		// Behandlung der Zeitdauer eines F0 Tastendrucks
		// NH kurz (<0,5s) NotHalt der eigenen Lok (<f "Register" "Adr" -1 "Richtung">)
		// NH lang (>0,5s) NotHalt aller Loks (<f "Register" "Adr" -1 1>)
		if (GETKEYNH && (NHTRIG == false))									// 	-> NH abfragen
		{
			NHTIME = millis();												// Startzeit des Tastendrucks merken
			NHTRIG = true;													// merken, dass getriggert wurde - Funktion nur einmalig ausloesen
			Serial.println("Nothalt kurz");										// Debug Ausgabe

			// Ausgabe an DCC++ BaseStation
			Serial.print("<t");												// lead in
			Serial.print(LOKREG); Serial.print(" ");						// Register
			Serial.print(LOKADR); Serial.print(" ");						// Lok Adresse
			Serial.print(-1); Serial.print(" ");							// Speed = -1 -> NOTHALT und anschliessend Speed = 0
			(LOKDIR) ? Serial.print("1") : Serial.print("0");				// Richtung (1=FWD, 0=REV)
			Serial.println(">");											// lead out

			NOTHALT = true;													// Status "NotHalt ausgeloest" fuer die aktuelle Lok merken
			LOKSPEED = 0;													// Fahrstufe der aktuellen Lok auf 0 setzen
		}
		else if ((NHTRIG == true) && (!GETKEYNH))							// Behandlung Taste F0 wieder losgelassen
		{
			NHTRIG = false;													// Trigger loeschen
			NHLONG = false;
		}

		if (GETKEYNH && (NHTRIG == true) && (NHLONG == false))				// langen Tastendruck behandeln     
		{
			if (millis() > (NHTIME + 500))									// NH laenger als 0,5s gedrueckt?
			{																//  -> NotHalt an ALLE Loks senden				
				uNotAusZeit = millis();										// Zeitpunkt "Notaus fuer ALLE" merken

				NHLONG = true;												// merken, dass LANG getriggert wurde - Funktion nur einmalig ausloesen
				Serial.println("Nothalt lang");									// Debug Ausgabe

				NotHaltALLE = true;											// Status merken fuer Ausgaberoutine
				NotHaltALLECounter = 0;										// Notaus ALLE wird mehrfach ausgegeen

				for (byte i = 0; i <= 7; i++)								// NotHalt Flag fuer alle Loks setzen, "i" ist die LokNummer
				{
					// NOTHALT = true;										// setzt das Flag nur fuer die _aktuelle_ Lok, daher Zugriff ueber das Array direkt
					LokArray[i][17] = true;									// hier liegt das NotHalt-Flag der Loks
					LokArray[i][7] = 0;										// hier liegt die Fahrstufe der Loks
				}
			}
		}

		// Der Befehl Nothalt an eine Lok wird von der BaseStation offenbar nur einmalig ausgegeben.
		// Wenn die betreffende Lok in diesem Moment gerade Kontaktprobleme auf den Gleisen hat, kommt die Abschaltung nicht an, die Lok faehrt unbeeindruckt weiter.
		// Also muss der Nothalt-Befehl sicherheitshalber mehrfach abgesetzt werden.
		if (NotHaltALLE == true)
		{
			if (millis() > uNotAusZeit + 200)								// "Notaus an ALLE" mehrmals im Abstand von 200 ms wiederholen
			{
				for (byte i = 0; i <= 7; i++)								// NotHalt Flag fuer alle Loks setzen, "i" ist die LokNummer
				{
					// Ausgabe an DCC++ BaseStation, alle Loks werden angehalten
					Serial.print("<t");										// lead in
					Serial.print(LokArray[i][0]); Serial.print(" ");		// Register
					Serial.print(LokArray[i][4]); Serial.print(" ");		// Lok Adresse
					Serial.print(-1); Serial.print(" ");					// Speed = 0 - NOTHALT
					Serial.print("1");										// Richtung (1=FWD, 0=REV)
					Serial.println(">");									// lead out
				}
				NotHaltALLECounter++;
				uNotAusZeit = millis();
			}

			if(NotHaltALLECounter > 3)										// die Nothalt Befehle werden 3 mal an die BaseStation gesendet
			{
				NotHaltALLE = false;
			}
		}
		
		// Behandlung der Zeitdauer eines F0 Tastendrucks					// F0 wird auch mit ausgeschaltetem Handset abgefragt und ausgewertet
		// Besondere Bedeutung: F0 lang (>1s) DCC-Spannung ein/aus (<1> / <0>)
		if (GETKEYF0 && (GETFxF0 == false))									//  -> F0 abfragen
		{
			F0TIME = millis();												// Startzeit des Tastendrucks merken
			SETFxF0(true);													// merken, dass getriggert wurde

			GETFxOldF0 == true ? SETFxOldF0(false) : SETFxOldF0(true);		// Toggle-Schalter Implementierung
			Serial.print("F0 pressed - "); Serial.println(GETFxOldF0);		// Debug Ausgabe
			//Serial.println("F0 kurz");									// Debug Ausgabe

			if (eepromReadInt(EEF0) == SCHALTER)							// ist F0 als Schalter parametriert?
			//if (GETTSF0 == SCHALTER)										// ist F0 als Schalter parametriert?
			//if ((FxOLD & 0x10 != 0x10))									// ist F0 als Schalter parametriert?
			{
				GETFxOldF0 == true ? functionCode += 16 : functionCode -= 16;	// F0 beim Druecken toggeln
				Serial.println("F0 Schalter");								// Debug Ausgabe
			}
			else
			{
				functionCode += 16;											// F0 ist als Taster parametriert
				Serial.println("F0 Taster");								// Debug Ausgabe
			}
			// Ausgabe an DCC++ BaseStation
			Serial.print("<f");												// lead in
			Serial.print(LOKADR); Serial.print(" ");						// Lok Adresse
			Serial.print(functionCode);										// Funktionscode
			Serial.println(">");											// lead out
		}
		else if ((GETFxF0 == true) && (!GETKEYF0))							// Behandlung Taste F0 wieder losgelassen
		{
			SETFxF0(false);													// Trigger loeschen
			F0LONG = false;

			//if ((FxOLD & 0x10 == 0x10))									// ist F0 als Schalter parametriert?
			//if (GETTSF0 == TASTER)										// ist F0 als Taster parametriert?
			if (eepromReadInt(EEF0) == TASTER)								// ist F0 als Taster parametriert?
			{
				functionCode -= 16;											// F0 inaktiv
			}
			// Ausgabe an DCC++ BaseStation
			Serial.print("<f");												// lead in
			Serial.print(LOKADR); Serial.print(" ");						// Lok Adresse
			Serial.print(functionCode);										// Funktionscode
			Serial.println(">");											// lead out
		}

		if (GETKEYF0 && (GETFxF0 == true) && (F0LONG == false))				// langen Tastendruck behandeln
		{
			if (millis() > (F0TIME + 1000))									// F0 laenger als 1s gedrueckt?
			{																//  -> DCC Spannung ein-/ausschalten
				F0LONG = true;
				Serial.println("F0 lang");									// Debug Ausgabe
				if (DCCSpannungEin == true)									// DCC eingeschaltet?
				{
					Serial.print("<0>");									// ..dann ausschalten mit langem Druck auf F0
					DCCSpannungEin = false;
				}
				else
				{
					Serial.print("<1>");									// Toggel-Funktion
					DCCSpannungEin = true;
				}
			}
		}

		if (SWVAROLD != SWVAR)												// hat sich an den Schaltern etwas geaendert?
		{
			// Debug Ausgaben
			//(GETHSON) ? Serial.print("Off - ") : Serial.print("On - ");	// invertierte Logik! Bit==1 -> Off
			//(LOKDIR) ? Serial.print("REV - ") : Serial.print("FWD - ");	// invertierte Logik! Bit==1 -> REV

			// Behandlung Modus Schalter - Anzahl Fahrstufen (zwei Steckbruecken auf 0x0C000, muss fuer DCC++ immer 3 sein, also Bruecken offen)
			MODUSBITS = (GETMODUS) >> 14;									// resultiert in Werten 0..3

			// da DCC++ ohnehin nur mit 128 Fahrstufen arbeitet, kann auf die switch-Konstruktion verzichtet werden
/*			switch (MODUSBITS)
			{
				// Debug Ausgaben
				case 0: Serial.println("not used - 1"); break;
				case 1: Serial.println("14 Speed Steps"); break;
				case 2: Serial.println("28 Speed Steps"); break;
				case 3: Serial.println("128 Speed Steps"); break;
				default: Serial.println("-- Wert undefiniert! --");
			}
*/
// ---------------------------------------------------------------------------------------------------------------------------------------
			// Lokadresse errechnen
			SWVARINT = SWVAR;												// Adresse auf 16 bit kuerzen fuer Berechnung der Lok-Adresse
			// am Handset eingestellte LokAdresse aus den Bits errechnen - die Schalter sind BCD-codiert, 2x 0..9 -> moegliche Adressen 0..99 - erlaubt 1..99
			LOKADR = ((((~(SWVARINT >> 1) & 0x00F0) >> 4) * 10) + (~(SWVARINT >> 1) & 0x000F));
			//Serial.println(LOKADR);										// DEBUG

// ---------------------------------------------------------------------------------------------------------------------------------------
			// Funktionstasten auswerten - das MK2 Handset hat 4 F-Tasten und eine Taste NotHalt
			//		<f CAB BYTE1 [BYTE2]> - Funktionstaste erfordert Function-Command <f...>
			// F0 = consist - bei mir NOTHALT	- F0..F3 = Funktionen
			//		To set functions F0 - F4 on(= 1) or off(= 0) :								// irrefuehrend! - es wird kein Parameter 0/1 mitgesendet
			//		BYTE1 : 128 + F0 * 1 + F1 * 2 + F2 * 4 + F3 * 8			(+ NH * 16)			// NH (Nothalt) wird an anderer Stelle ausgewertet
			//		BYTE2 : omitted			

			if (GETHSON == true)											// Funktionstasten nur an die BaseStation ausgeben wenn Handset "On"
			{
#ifndef TIMINGCHECK
				digitalWrite(SignalLED, 1);	LEDmillis = millis();			// Arduino LED fuer 100 ms einschalten - Anzeige, dass Daten an die BaseStation uebertragen werden
#endif
				// F-taste F0 abfragen												// hier nicht mehr abfragen, F0-Taste wird oben schon abgefragt (inklusive "langer Druck")

				// F-taste F1 abfragen
				if (GETKEYF1 && (GETFxF1 == false))									//  -> F1 abfragen
				{
					SETFxF1(true);													// merken, dass getriggert wurde

					GETFxOldF1 == true ?  SETFxOldF1(false) : SETFxOldF1(true);		// Toggle-Schalter Implementierung
					Serial.print("F1 pressed - "); Serial.println(GETFxOldF1);		// Debug Ausgabe

					//if (GETTSF1 == SCHALTER)										// ist F1 als Schalter parametriert?
					//if ((FxOLD & 0x20 == 0x00))									// ist F1 als Schalter parametriert?
					if (eepromReadInt(EEF1) == SCHALTER)							// ist F1 als Schalter parametriert?
					{
						GETFxOldF1 == true ? functionCode += 2 : functionCode -= 2;	// F1 beim Druecken toggeln
					}
					else
					{
						functionCode += 2;											// F1 ist als Taster parametriert
					}
					// Ausgabe an DCC++ BaseStation
					Serial.print("<f");												// lead in
					Serial.print(LOKADR); Serial.print(" ");						// Lok Adresse
					Serial.print(functionCode);										// Funktionscode
					Serial.println(">");											// lead out
				}
				else if ((GETFxF1 == true) && (!GETKEYF1))							// Behandlung Taste F1 wieder losgelassen
				{
					SETFxF1(false);													// Trigger loeschen

					//if ((FxOLD & 0x20 == 0x20))									// ist F1 als Taster parametriert?
					//if (GETTSF1 == TASTER)										// ist F1 als Taster parametriert?
					if (eepromReadInt(EEF1) == TASTER)								// ist F1 als Taster parametriert?
					{
						functionCode -= 2;											// F1 inaktiv
					}
					// Ausgabe an DCC++ BaseStation
					Serial.print("<f");												// lead in
					Serial.print(LOKADR); Serial.print(" ");						// Lok Adresse
					Serial.print(functionCode);										// Funktionscode
					Serial.println(">");											// lead out
				}

				// F-taste F2 abfragen
				if (GETKEYF2 && (GETFxF2 == false))									//  -> F2 abfragen
				{
					SETFxF2(true);													// merken, dass getriggert wurde

					GETFxOldF2 == true ? SETFxOldF2(false) : SETFxOldF2(true);		// Toggle-Schalter Implementierung
					Serial.print("F2 pressed - "); Serial.println(GETFxOldF2);		// Debug Ausgabe

					if (eepromReadInt(EEF2) == SCHALTER)							// ist F2 als Schalter parametriert?
					{
						GETFxOldF2 == true ? functionCode += 4 : functionCode -= 4;	// F2 beim Druecken toggeln
					}
					else
					{
						functionCode += 4;											// F2 ist als Taster parametriert
					}
					// Ausgabe an DCC++ BaseStation
					Serial.print("<f");												// lead in
					Serial.print(LOKADR); Serial.print(" ");						// Lok Adresse
					Serial.print(functionCode);										// Funktionscode
					Serial.println(">");											// lead out
				}
				else if ((GETFxF2 == true) && (!GETKEYF2))							// Behandlung Taste F2 wieder losgelassen
				{
					SETFxF2(false);													// Trigger loeschen

					if (eepromReadInt(EEF2) == TASTER)								// ist F2 als Taster parametriert?
					{
						functionCode -= 4;											// F2 inaktiv
					}
					// Ausgabe an DCC++ BaseStation
					Serial.print("<f");												// lead in
					Serial.print(LOKADR); Serial.print(" ");						// Lok Adresse
					Serial.print(functionCode);										// Funktionscode
					Serial.println(">");											// lead out
				}

				// F-taste F3 abfragen
				if (GETKEYF3 && (GETFxF3 == false))									//  -> F3 abfragen
				{
					SETFxF3(true);													// merken, dass getriggert wurde

					GETFxOldF3 == true ? SETFxOldF3(false) : SETFxOldF3(true);		// Toggle-Schalter Implementierung
					Serial.print("F3 pressed - "); Serial.println(GETFxOldF3);		// Debug Ausgabe

					if (eepromReadInt(EEF3) == SCHALTER)							// ist F3 als Schalter parametriert?
					{
						GETFxOldF3 == true ? functionCode += 8 : functionCode -= 8;	// F3 beim Druecken toggeln
					}
					else
					{
						functionCode += 8;											// F3 ist als Taster parametriert
					}
					// Ausgabe an DCC++ BaseStation
					Serial.print("<f");												// lead in
					Serial.print(LOKADR); Serial.print(" ");						// Lok Adresse
					Serial.print(functionCode);										// Funktionscode
					Serial.println(">");											// lead out
				}
				else if ((GETFxF3 == true) && (!GETKEYF3))							// Behandlung Taste F3 wieder losgelassen
				{
					SETFxF3(false);													// Trigger loeschen

					if (eepromReadInt(EEF3) == TASTER)								// ist F3 als Taster parametriert?
					{
						functionCode -= 8;											// F3 inaktiv
					}
					// Ausgabe an DCC++ BaseStation
					Serial.print("<f");												// lead in
					Serial.print(LOKADR); Serial.print(" ");						// Lok Adresse
					Serial.print(functionCode);										// Funktionscode
					Serial.println(">");											// lead out
				}
			}
			else
			{		// hier ist der Handset Schalter auf OFF, es werden keine Signale dieses Handset an die BaseStation	gesandt
				if (LOKADR != 0)	// allerdings nur, wenn das entsprechende Handset eine Lokadresse ungleich Null eingestellt hat,
									// sonst wird beim Einschalten fr alle nicht angeschlossenen Handsets ein "Speed = 0" abgesetzt
									// was die BaseStation dazu animiert, fortan diese unnuetzen Befehle zu senden
				{
					// Ausgabe an DCC++ BaseStation
					Serial.print("<t");										// lead in
					Serial.print(LOKREG); Serial.print(" ");				// Register
					Serial.print(LOKADR); Serial.print(" ");				// Lok Adresse
					Serial.print(0); Serial.print(" ");						// Speed = 0 mit Abbremsen
					(LOKDIR) ? Serial.print("1") : Serial.print("0");		// Richtung (1=FWD, 0=REV)
					Serial.println(">");									// lead out
				}

				NOTHALT = true;			// Status "NotHalt ausgeloest" -> Handset war ausgeschaltet, daher darf die Lok erst wieder fahren, wenn Speed = 0 eingestellt war 
			}

			SWVAROLD = SWVAR;												// aktuelle Schalterstellung je Lok merken fuer Abfrage ob geaendert
		}

// ---------------------------------------------------------------------------------------------------------------------------------------
		// hier die Potis auswerten ------------------------------
		LOKPOTI = (((n - 1) * LOKPOTI) + analogRead(MK2POTIIN)) / n;		// exponentieller Mittelwert gegen Zappeln der Werte (n siehe Anfang von loop() )

		// die nachfolgende switch-Anweisung verhindert irgendwie (ungeklaert) die korrekte Funktion von SoftwareSerial mit Pin 11 als Ausgang
		// da DCC++ ohnehin nur mit 128 Fahrstufen arbeitet, kann auf die switch-Konstruktion verzichtet werden (koennte jetzt wieder aktiviert werden, SOftSerial ist raus)
/*		switch (MODUSBITS)	// je nach Modus-Bits im MK2 die Potistellung auf die Fahrstufen / SpeedSteps mappen (fuer DCC++ IMMER 128 auswaehlen!!)
		{
			case 0: Serial.println("not used - 2"); 	break;
			case 1: LOKSPEED = map(LOKPOTI, 4, 1010, 0, 14); LOKSPEED = constrain(LOKSPEED, 0, 14); break;
			case 2:	LOKSPEED = map(LOKPOTI, 4, 1010, 0, 28); LOKSPEED = constrain(LOKSPEED, 0, 28); break;
			case 3:	LOKSPEED = map(LOKPOTI, 4, 1010, 0, 126); LOKSPEED = constrain(LOKSPEED, 0, 126); break;	// "126" ist Max fuer DCC++ BaseStation
			default: Serial.println("- - - Wert undefiniert! - - -"); break;
		}
*/
		LOKSPEED = map(LOKPOTI, 4, 1010, 0, 126);							// "126" ist Max fuer DCC++ BaseStation
		LOKSPEED = constrain(LOKSPEED, 0, 126);								// auf die Grenzen beschraenken

		if (LOKSPEED == 0) NOTHALT = false;									// nach NotHalt muss Speed = 0 eingestellt werden, bevor eine Lok wieder fahren darf

		if ((LOKSPEED != LOKSPEEDOLD) || (LOKDIR !=LOKDIROLD))				// hat sich die Fahrstufe oder die Fahrtrichtung geaendert?
		{
			//Serial.print(" - map Wert: ");  Serial.println(LOKSPEED]);		// DEBUG
			//Serial.print("LokNummer: "); Serial.print(LokNummer);  Serial.print("Speed: "); Serial.print(LOKPOTI); // DEBUG

			//		<t REGISTER CAB SPEED DIRECTION> - Fahrstufenaenderung erfordert Throttle-Command <t...>
			if((GETHSON == true) && (NOTHALT == false))						// nur ausgeben wenn Handset "On" und kein NotHalt ausgeloest ist
			{
#ifndef TIMINGCHECK
				digitalWrite(SignalLED, 1);	LEDmillis = millis();			// Arduino LED fuer 100 ms einschalten - Anzeige, dass Daten an die BaseStation uebertragen werden
#endif
				// Ausgabe an DCC++ BaseStation
				Serial.print("<t");											// lead in
				Serial.print(LOKREG); Serial.print(" ");					// Register
				Serial.print(LOKADR); Serial.print(" ");					// Lok Adresse
				Serial.print(LOKSPEED); Serial.print(" ");					// Speed
				(LOKDIR) ? Serial.print("1") : Serial.print("0");			// Richtung (1=FWD, 0=REV)
				Serial.println(">");										// lead out
			}
			LOKSPEEDOLD = LOKSPEED;											// neuer Alt-Wert fuer Vergleich auf Aenderung
			LOKDIROLD = LOKDIR;												// neuer Alt-Wert fuer Vergleich auf Aenderung
		}
	}

	// Parameter verndern oder im EEPROM speichern oder dort holen --------------------------------------------------------------
	if (stringComplete == true)
	{
		if (inputString.indexOf("$$") >= 0) 					// vorhandene Werte abfragen/ausgeben
		{
			Serial.print(F("Version ")); Serial.print(MajorVersion); Serial.print("."); Serial.print(MinorVersion, 1); Serial.print(", "); Serial.print(BuildDate); Serial.println(F(", Harald Sattler"));
			Serial.println();
			Serial.println(F("F-Tasten arbeiten als Schalter oder Taster")); 
			Serial.println(F("Die Einstellung gilt immer fuer ALLE Handsets")); 
			Serial.println();
			Serial.print(F("F0 ist ")); Serial.println((eepromReadInt(EEF0) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F1 ist ")); Serial.println((eepromReadInt(EEF1) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F2 ist ")); Serial.println((eepromReadInt(EEF2) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F3 ist ")); Serial.println((eepromReadInt(EEF3) == TASTER ? F("Taster") : F("Schalter")));
			Serial.println();
		} else
		if (inputString.indexOf("F0") >= 0) 					// Taste F0 parametrieren
		{
			if (inputString.indexOf("T") >= 0) 					// Taste F0 als Taster definieren
			{
				eepromWriteInt(EEF0, TASTER);
				SETTSF0(true);									// Taster -> true
			}
			else if (inputString.indexOf("S") >= 0)				// Taste F0 als Schalter definieren
			{
				eepromWriteInt(EEF0, SCHALTER);
				SETTSF0(false);									// Schalter -> false
				SETFxOldF0(false);								// Toggle-Merker auf false, damit beim naechsten Druck auf Taster EIN-geschaltet wird
			}
			Serial.print(F("F0 ist ")); Serial.println((eepromReadInt(EEF0) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F1 ist ")); Serial.println((eepromReadInt(EEF1) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F2 ist ")); Serial.println((eepromReadInt(EEF2) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F3 ist ")); Serial.println((eepromReadInt(EEF3) == TASTER ? F("Taster") : F("Schalter")));
			Serial.println();
		} else
		if (inputString.indexOf("F1") >= 0) 					// Taste F1 parametrieren
		{
			if (inputString.indexOf("T") >= 0) 					// Taste F1 als Taster definieren
			{
				eepromWriteInt(EEF1, TASTER);
				SETTSF1(true);									// Taster -> true
			}
			else if (inputString.indexOf("S") >= 0)				// Taste F1 als Schalter definieren
			{
				eepromWriteInt(EEF1, SCHALTER);
				SETTSF1(false);									// Schalter -> false
				SETFxOldF1(false);								// Toggle-Merker auf false, damit beim naechsten Druck auf Taster EIN-geschaltet wird
			}
			Serial.print(F("F0 ist ")); Serial.println((eepromReadInt(EEF0) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F1 ist ")); Serial.println((eepromReadInt(EEF1) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F2 ist ")); Serial.println((eepromReadInt(EEF2) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F3 ist ")); Serial.println((eepromReadInt(EEF3) == TASTER ? F("Taster") : F("Schalter")));
			Serial.println();
		} else
		if (inputString.indexOf("F2") >= 0) 					// Taste F2 parametrieren
		{
			if (inputString.indexOf("T") >= 0) 					// Taste F2 als Taster definieren
			{
				eepromWriteInt(EEF2, TASTER);
				SETTSF2(true);									// Taster -> true
			}
			else if (inputString.indexOf("S") >= 0)				// Taste F2 als Schalter definieren
			{
				eepromWriteInt(EEF2, SCHALTER);
				SETTSF2(false);									// Schalter -> false
				SETFxOldF2(false);								// Toggle-Merker auf false, damit beim naechsten Druck auf Taster EIN-geschaltet wird
			}

			Serial.print(F("F0 ist ")); Serial.println((eepromReadInt(EEF0) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F1 ist ")); Serial.println((eepromReadInt(EEF1) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F2 ist ")); Serial.println((eepromReadInt(EEF2) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F3 ist ")); Serial.println((eepromReadInt(EEF3) == TASTER ? F("Taster") : F("Schalter")));
			Serial.println();
		} else
		if (inputString.indexOf("F3") >= 0) 					// Taste F3 parametrieren
		{
			if (inputString.indexOf("T") >= 0) 					// Taste F3 als Taster definieren
			{
				eepromWriteInt(EEF3, TASTER);
				SETTSF3(true);									// Taster -> true
			}
			else if (inputString.indexOf("S") >= 0)				// Taste F3 als Schalter definieren
			{
				eepromWriteInt(EEF3, SCHALTER);
				SETTSF3(false);									// Schalter -> false
				SETFxOldF3(false);								// Toggle-Merker auf false, damit beim naechsten Druck auf Taster EIN-geschaltet wird
			}

			Serial.print(F("F0 ist ")); Serial.println((eepromReadInt(EEF0) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F1 ist ")); Serial.println((eepromReadInt(EEF1) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F2 ist ")); Serial.println((eepromReadInt(EEF2) == TASTER ? F("Taster") : F("Schalter")));
			Serial.print(F("F3 ist ")); Serial.println((eepromReadInt(EEF3) == TASTER ? F("Taster") : F("Schalter")));
			Serial.println();
		} else
		{
			Serial.println();
			Serial.println(F("Gueltige Kommandos sind '$$', 'Fx=T', 'Fx=S' mit x[0..3]"));
			Serial.println();
		}

		// nach Umparametrierung der F-Tasten muss definiert neu aufgesetzt werden
		functionCode = 128;														// der Taster-Code ist immer mindestens 128, je nach aktiver F-Taste wird er erhoeht

		// clear the string
		inputString = "";
		stringComplete = false;
	}
	// ENDE - Parameter verndern oder im EEPROM speichern oder dort holen -------------------------------------------------------

#ifdef TIMINGCHECK
	//digitalWrite(SignalLED, !digitalRead(SignalLED));							// toggle LED fuer Laufzeitmessung
	digitalWrite(SignalLED, LOW);												// LED aus fuer Laufzeitmessung
#endif
}
// loop() Ende ---------------------------------------------------------------------------------------------------------------------------

// ---------------------------------------------------------------------------------------------------------------------------------------
 // Abfrage der seriellen Schnittstelle
void serialEvent() {
	while (Serial.available()) {
		// get the new byte:
		char inChar = (char)Serial.read();
		// add it to the inputString:
		inputString += inChar;
		// if the incoming character is a newline, set a flag
		// so the main loop can do something about it:
		if (inChar == '\r') {
			stringComplete = true;
		}
	}
}

// Hilfsroutinen Integer in/aus EEPROM
void eepromWriteInt(int adr, int wert) {
	byte low, high;
	low = wert & 0xFF;
	high = (wert >> 8) & 0xFF;
	EEPROM.write(adr, low); // dauert 3,3ms
	EEPROM.write(adr + 1, high);
	return;
} //eepromWriteInt

int eepromReadInt(int adr) {
	byte low, high;
	low = EEPROM.read(adr);
	high = EEPROM.read(adr + 1);
	return low + ((high << 8) & 0xFF00);
} //eepromReadInt

